"""This module contains the mosaik meta definition for the
pandapower simulator.

"""

META = {
    "type": "time-based",
    "models": {
        "Grid": {
            "public": True,
            "params": [
                "gridfile",  # Name of the grid topology
                "grid_params",
                "plotting",  # Flag to activate plotting
                "use_constraints",  # Flag to activate constraints
                "constraints",  # A list of constraints to activate
            ],
            "attrs": ["health", "grid_json"],
        },
        "Ext_grid": {
            "public": False,
            "params": [],
            "attrs": [
                "p_mw",  # load active power [MW]
                "q_mvar",  # load reactive power [MVAr]
            ],
        },
        "Bus": {
            "public": False,
            "params": [],
            "attrs": [
                "p_mw",  # load Active power [MW]
                "q_mvar",  # Reactive power [MVAr]
                "vn_kv",  # Nominal bus voltage [KV]
                "vm_pu",  # Voltage magnitude [p.u]
                "va_degree",  # Voltage angle [deg]
            ],
        },
        "Load": {
            "public": False,
            "params": [],
            "attrs": [
                "p_mw",  # load Active power [MW]
                "q_mvar",  # Reactive power [MVAr]
                "in_service",  # specifies if the load is in service.
                "controllable",  # States if load is controllable or not.
            ],
        },
        "Sgen": {
            "public": False,
            "params": [],
            "attrs": [
                "p_mw",  # load Active power [MW]
                "q_mvar",  # Reactive power [MVAr]
                "in_service",  # specifies if the load is in service.
                "controllable",  # States if load is controllable or not.
                "va_degree",  # Voltage angle [deg]
            ],
        },
        "Trafo": {
            "public": False,
            "params": [],
            "attrs": [
                "p_hv_mw",  # Active power at "from" side [MW]
                "q_hv_mvar",  # Reactive power at "from" side [MVAr]
                "p_lv_mw",  # Active power at "to" side [MW]
                "q_lv_mvar",  # Reactive power at "to" side [MVAr]
                "sn_mva",  # Rated apparent power [MVA]
                "max_loading_percent",  # Maximum Loading
                "vn_hv_kv",  # Nominal primary voltage [kV]
                "vn_lv_kv",  # Nominal secondary voltage [kV]
                "pl_mw",  # Active power loss [MW]
                "ql_mvar",  # reactive power consumption of the
                # transformer [Mvar]
                # 'pfe_kw',       #  iron losses in kW [kW]
                # 'i0_percent',       #  iron losses in kW [kW]
                "loading_percent",  # load utilization relative to rated
                # power [%]
                "i_hv_ka",  # current at the high voltage side of the
                # transformer [kA]
                "i_lv_ka",  # current at the low voltage side of the
                # transformer [kA]
                "tap_max",  # maximum possible  tap turns
                "tap_min",  # minimum possible tap turns
                "tap_pos",  # Currently active tap turn
            ],
        },
        "Line": {
            "public": False,
            "params": [],
            "attrs": [
                "p_from_mw",  # Active power at "from" side [MW]
                "q_from_mvar",  # Reactive power at "from" side [MVAr]
                "p_to_mw",  # Active power at "to" side [MW]
                "q_to_mvar",  # Reactive power at "to" side [MVAr]
                "max_i_ka",  # Maximum current [KA]
                "length_km",  # Line length [km]
                "pl_mw",  # active power losses of the line [MW]
                "ql_mvar",  # reactive power consumption of the line [MVar]
                "i_from_ka",  # Current at from bus [kA]
                "i_to_ka",  # Current at to bus [kA]
                "loading_percent",  # line loading [%]
                "r_ohm_per_km",  # Resistance per unit length [Ω/km]
                "x_ohm_per_km",  # Reactance per unit length [Ω/km]
                "c_nf_per_km",  # Capactity per unit length [nF/km]
                "in_service",  # Boolean flag (True|False)
            ],
        },
        "Switch": {
            "public": False,
            "params": [],
            "attrs": [
                "et",
                "type",
                "closed",
            ],
        },
        "Storage": {
            "public": False,
            "params": [],
            "attrs": [
                "p_mw",  # load Active power [MW]
                "q_mvar",  # Reactive power [MVAr]
                "max_e_mwh"  # maximum charge level
                "in_service",  # specifies if the load is in service.
                "controllable",  # States if load is controllable or not.
            ],
        },
    },
}



from typing import List

from ..elements.bus import PPBus
from ..elements.line import PPLine
from ..elements.load import PPLoad
from ..elements.sgen import PPSgen
from ..elements.transformer import PPTransformer
from . import LOG
from .static import PandapowerGrid


class ConstraintedGrid(PandapowerGrid):
    def __init__(self, constraints):
        super().__init__()

        self._constraints_to_load = constraints
        self.constraints = dict()
        self._lf_states: List[bool] = []  # just for more precise logging

    def setup(self, gridfile, grid_idx, grid_params):
        super().setup(gridfile, grid_idx, grid_params)

        self._load_constraints()

    def set_inputs(self, etype, idx, data):
        etype = etype.lower()
        if etype not in ["load", "sgen", "trafo", "switch", "storage"]:
            LOG.info("Invalid etype %s. Skipping.", etype)
            return False

        for name, value in data.items():
            self.grid[etype].at[idx, name] = value

            if etype in self.constraints:
                # Constraint can change the value
                setattr(self.constraints[etype][idx], name, value)

    def run_powerflow(self, time, max_iter=2):

        self._lf_states.append(self.lf_converged)  # previous state

        # Run once to check current state
        super().run_powerflow(time)

        if time < 0:
            # The first step is done before the simulation to make use
            # of numba's speed-up during the simulation
            return

        state_changed = False

        # Now constraints can change the input state if necessary
        # Constraints will definitively change the state of the grid
        # so another power flow calculation is required afterwards
        for key in ["trafo", "load", "sgen", "line", "bus"]:
            if key not in self.constraints:
                continue
            for element in self.constraints[key]:
                state_changed = element.step(time) or state_changed

        # Maybe more elements were put out of service than necessary
        # e.g., when the critical element is the last one checked.
        # Performing another iteration allows elements to switch on
        # again if no constraints are violated.
        if max_iter > 1:
            return self.run_powerflow(time, max_iter - 1)

        super().run_powerflow(time)
        if state_changed:
            if not self._lf_states[0] and self.lf_converged:
                LOG.info(f"At step {time}: Constraints fixed failing LF.")
            if not self._lf_states[0] and not self.lf_converged:
                LOG.info(f"At step {time}: LF still not converging.")
            if self._lf_states[0] and not self.lf_converged:
                LOG.info(f"At step {time}: LF broke due to constraints.")
            if self._lf_states[0] and self.lf_converged:
                LOG.info(f"At step {time}: LF still converging.")

        self._lf_states = []

    def _load_constraints(self):
        for constr in self._constraints_to_load:
            etype, value = constr
            self.constraints.setdefault(etype, list())
            for idx in range(len(self.grid[etype])):
                self.constraints[etype].append(self._create(etype, idx, value))

    def _create(self, etype, index, value):
        if etype == "trafo":
            clazz = PPTransformer
        # if classname == PPTransformer:
        #     etype = self.grid[classname.pp_key()]["std_type"][index]

        #     return PPTransformer(index, self.grid)
        # TODO: elif other elements
        elif etype == "bus":
            clazz = PPBus
        elif etype == "load":
            clazz = PPLoad
        elif etype == "sgen":
            clazz = PPSgen
        elif etype == "line":
            clazz = PPLine
        return clazz(index, self.grid, value)
